/*
 * patch/hints for Relentless/Little Big Adventure
 * by Aaron Wohl (n3liw+@cmu.edu)
 * compiled with borland c++ 4.5
 */

#include <iostream.h>
#include <fstream.h>
#include <strstrea.h>
#include <dirent.h>
#include <ctype.h>
#include <assert.h>
#include <cstring.h>
#include "classlib\sets.h"

unsigned      _RTLENTRY _stklen=30000;

#ifndef TRUE
#define TRUE (0==0)
#endif
#ifndef FALSE
#define FALSE (0==1)
#endif

// representation of a game
#define LBA_DATA_SIZE (459)
class lba;
class var_bool;
class var_byte;

class lba {
private:
	string fname;
	string uname;
	unsigned char raw_bytes[LBA_DATA_SIZE];	//the saved game info

public:
	static TSet<lba*> saved_games;
	friend lba *find_lba_by_name(const char *name);

	void put_byte(int voffset,int new_byte) {raw_bytes[voffset]=new_byte;}
        int get_byte(int voffset) { return int(raw_bytes[voffset])&0x0FF;}

	const char *get_name(void) { return uname.c_str(); };
	lba(const char *fname);		//read in this file
	void diff_lba(lba &other);	//print differences with other guy
	int match_name(const char *name);
	void write_lba(void);		//write yourself out
};

TSet<lba*> lba::saved_games;

// abstract type for a variable of any time
class var_info {
protected:
	string vname;
	int voffset;
public:
	var_info(int aoffset,const char *aname)
			: vname(aname),voffset(aoffset)
			{ all_vars[aoffset]=this;
			  vname.set_case_sensitive(FALSE);
			};
	static var_info* all_vars[LBA_DATA_SIZE];
	ostream &print_name(ostream &co) { co << vname; return co;}
	friend var_info *find_var_by_name(const char *aname);

        // print the value of this variable in the passed lba
	virtual void print_value(lba &)=0;
	// set a new value for the this variable in the passed lba
	virtual void set_value(lba &alba,const char *newval)=0;

	// return size in bytes take up in save file
	virtual int get_size(void) { return 1; }

	//after setting value print out its new value
	virtual void post_set_value(lba &alba) {
		print_name(cout);
		cout << " new value ";
		print_value(alba);
		cout << endl;
	}
};
var_info* var_info::all_vars[LBA_DATA_SIZE];

// find a variable by its name
var_info *find_var_by_name(const char *aname)
{
  for(int i=0;i<LBA_DATA_SIZE;i++)
   if(var_info::all_vars[i]->vname==aname)
	   return var_info::all_vars[i];
  cerr << "No such variable as " << aname << endl;
  exit(1);
  return 0;
}

// print a nice name for a variable
// ostream &print_voname(ostream &co,int voffset)
ostream _FAR & _RTLENTRY _EXPFUNC print_voname(ostream _FAR &co,int voffset)
{
  return var_info::all_vars[voffset]->print_name(co);
}

// a boolean variable
class var_bool : public var_info {
public:
  var_bool(int aoffset,const char *aname) : var_info(aoffset,aname) { };
  virtual void print_value(lba &alba);
  virtual void set_value(lba &alba,const char *newval);
};

// set a new value
void var_bool::set_value(lba &alba,const char *newval)
{
  char ch=newval[0];
  if(strlen(newval)!=1) {
   cerr << "Valid values for boolean variables are t and f" << endl;
   exit(1);
  }
  if(isupper(ch))
    ch=tolower(ch);
  if(ch=='t')
     alba.put_byte(voffset,1);
  else if (ch=='f')
     alba.put_byte(voffset,0);
  else {
     cerr << "Valid values for boolean variables are t and f" << endl;
     exit(1);
  }
  post_set_value(alba);
}

// print the text representation of this variable
void var_bool::print_value(lba &alba)
{
  if(alba.get_byte(voffset))
    cout << "true";
  else
    cout << "false";
}

class var_word : public var_info {
public:
  var_word(int aoffset,const char *aname) : var_info(aoffset,aname) { };
  virtual void print_value(lba &alba) 
    { int result=alba.get_byte(voffset+1)*256+alba.get_byte(voffset);
      cout << result;
    }
  virtual void set_value(lba &alba,const char *newval);
  virtual int get_size(void) { return 2; }
};

// set a new value
void var_word::set_value(lba &alba,const char *newval)
{
  int aval=atol(newval);
  alba.put_byte(voffset,aval);
  alba.put_byte(voffset+1,aval>>8);
  post_set_value(alba);
}

class var_byte : public var_info {
public:
  var_byte(int aoffset,const char *aname) : var_info(aoffset,aname) { };
  virtual void print_value(lba &alba) 
    { cout << alba.get_byte(voffset);}
  virtual void set_value(lba &alba,const char *newval);
};

// set a new value
void var_byte::set_value(lba &alba,const char *newval)
{
  int aval=atol(newval);
  alba.put_byte(voffset,aval);
  post_set_value(alba);
}

// add variables
static void add_vars(void)
{
  new var_bool(2,"holomap");
  new var_bool(3,"ball");
  new var_bool(4,"sword");
  new var_bool(5,"horn");
  new var_bool(6,"tunic");
  new var_bool(7,"book");
  new var_bool(8,"medalion");
  new var_bool(9,"water");
  new var_bool(10,"red_card");
  new var_bool(11,"blue_card");
  new var_bool(12,"my_id");
  new var_bool(13,"architects_id");
  new var_bool(14,"jet_pack");
  new var_bool(15,"ski_board");
  new var_bool(16,"penquin");
  new var_bool(17,"gas_can");
  new var_bool(18,"flag");
  new var_bool(19,"flute");
  new var_bool(20,"guitar");
  new var_bool(21,"hair_dryer");
  new var_bool(22,"acestral_key");
  new var_bool(23,"syrup");
  new var_bool(24,"empty_bottle");
  new var_bool(25,"ferry_ticket");
  new var_bool(26,"keypad");
  new var_bool(27,"coffie_pot");
  new var_bool(28,"bonus_list");

  new var_bool(29,"opened_citadel_chest");
  new var_bool(94,"bought_citadel_grobo_drink");

  new var_byte(257,"scean");	//where in the game we are

  new var_byte(456,"Clovers");	//max is number of clovers
  new var_byte(260,"Health");	//max 50
  new var_word(261,"Kashes");	//max 999 I heard

  for(int i=0;i<LBA_DATA_SIZE;i++)
   if(var_info::all_vars[i]==0) {
     ostrstream dname;	   
     dname << "V" << i << ends;
     new var_byte(i,dname.str());
   }
}

void xx_fake_refrence(void *foo)
{
}

#define FAKE_REFRENCE(xx_arg) xx_fake_refrence((void*)xx_arg)

// see if our name matches the passed name
int lba::match_name(const char *name)
{
  return (uname.compare((char*)name)==0);
}

// see if this lba matches the passed name
static int match_name(lba * const &an_item,void *name)
{
  return an_item->match_name((char*)name);
}

// search for an lba with the passed name
lba *find_lba_by_name(const char *name)
{
  lba **pt=lba::saved_games.FirstThat(match_name,(void*)name);
  if(pt==0)
	  return 0;
  return *pt;
}

// search for an lba with the passed name
// fail if it can't be found
lba *must_find_lba_by_name(const char *name)
{
  lba *result=find_lba_by_name(name);
  if(result==0) {
     cerr << "Can't find game '" << name << "'" << endl;
     exit(1);
  }
  return result;
}
	      
// mask a number to it's low byte
static inline int INT1(int xx_arg)
{
	return xx_arg&0x0FF;
}

//print differences with other guy
void lba::diff_lba(lba &other)
{
  cout << "Differences between " << uname << " and " << other.uname << endl;
  for(int i=0;i<LBA_DATA_SIZE;) {
    var_info *vi=var_info::all_vars[i];
    if(raw_bytes[i]!=other.raw_bytes[i]) {
      print_voname(cout,i);
      cout<< " ";
      vi->print_value(*this);
      cout<< " ";
      vi->print_value(other);
      cout << endl;
    }




    i+=vi->get_size();
  }
}

// write out the current this object
void lba::write_lba(void)
{
  ofstream writeme(fname.c_str());
  writeme << char(3);
  writeme << uname;
  writeme.write(raw_bytes,LBA_DATA_SIZE);
  if(writeme.fail())
	  cout << "write failed" << endl;
}

// read in the named file
lba::lba(const char *afname) : fname(afname)
{
  uname.set_case_sensitive(FALSE);
  ifstream readme(fname.c_str(),ios::in);
  unsigned char fname_len;
  readme >> fname_len;
  char buf[512];
  fname_len&=0x0FF;
  readme.get(buf,sizeof(buf)-1,0);
  buf[sizeof(buf)-1]=0;
  fname_len=strlen(buf);
  uname=buf;
//  cout << "Loaded \"" << uname << "\"" << endl;
  readme.get(raw_bytes,LBA_DATA_SIZE);
  if(readme.fail())
	  cout << "read failed" << endl;
  if(readme.eof())
	  cout << "unexpected eof" << endl;
  saved_games.Add(this);
}

// is this a lba save file name?
static int is_save_file_name(const char *fname)
{
  char ch= *fname++;
  if(ch!='S')
    return FALSE;
  while((ch=(*fname++))!=0) {
    if(!isdigit(ch))
	    break;
  }
  if(ch!='.')
    return FALSE;
  if(memcmp(fname,"LBA",4)!=0)
   return FALSE;
  return TRUE;
}

// called on each set item to count it
static void count_set_item(lba *&an_item,void *count)
{
  FAKE_REFRENCE(an_item);
  long *lcount=(long*)count;
  (*lcount)++;
}

// return the number of elements in a set
long get_set_size(TSet<lba*> &aset)
{
  long set_size=0;
  aset.ForEach(count_set_item,&set_size);
  return set_size;
}

// load the lba files and read in the names
static void scan_dir(void)
{
  DIR _FAR *adir=opendir(".");
  for(struct dirent _FAR *afil=readdir(adir);afil!=0;afil=readdir(adir)) {
    if(!is_save_file_name(afil->d_name))
	    continue;
    new lba(afil->d_name);
  }
  cout << "Loaded " <<
		  get_set_size(lba::saved_games) << " saved games " << endl;
}

// say how to use the program and quit
static void usage(void)
{
  cout << "run patlba in your relentless directory" << endl;
  cout << "patlba list - shows saved files" << endl;
  cout << "patlba patch SAVE_NAME VARIABLE VALUE" << endl;
  cout << "patlba diff SAVE_NAME1 SAVE_NAME2" << endl;
  cout << "examples:" << endl;
  cout << "  patlba diff c1 c2" << endl;
  cout << "  patlba patch gamex clovers 5" << endl;
  cout << "  patlba patch gamex kashes 600" << endl;
  cout << "  patlba patch gamey red_card t" << endl;
  cout << "  WARNING: BACKUP YOUR GAME FILES BEFORE PATCHING THEM" << endl;
  exit(1);
}

// called on each saved game to list it
static void list_name(lba *&an_item,void *rubbish)
{
  FAKE_REFRENCE(rubbish);
  cout << an_item->get_name() << " ";
}

// list the available games
static void list_games(void)
{
  cout << "Available saved games:" << endl;
  lba::saved_games.ForEach(list_name,0L);
}

// are the two passed strings equal?
inline int streql(const char *s1,const char *s2)
{
  return (strcmp(s1,s2)==0);
}

// compare two games and print the differences
static void diff_games(const char *g1_name,const char *g2_name)
{
  lba *g1=must_find_lba_by_name(g1_name);
  lba *g2=must_find_lba_by_name(g2_name);
  g1->diff_lba(*g2);
}

// patch the passed game
static void patch_game(const char *game_name,const char *var,const char *newval)
{
  lba *game=must_find_lba_by_name(game_name);
  var_info *vi=find_var_by_name(var);
  cout << "Patching " << var << " of "
		  << game->get_name() << " to " << newval << endl;
  vi->set_value(*game,newval);
  game->write_lba();
}

// parse and execute a command
static void parse_cmd(char *cmd,char **args,int arg_count)
{
  if(streql(cmd,"list"))
    list_games();
  else if(streql(cmd,"diff")) {
    if(arg_count!=2) {
      cerr << "patlba diff needs two save game names as arguments" << endl;
      exit(1);
    }	    
    diff_games(args[0],args[1]);
  } else if(streql(cmd,"patch")) {
    if(arg_count!=3) {
      cerr << "patlba needs the game, variable to patch and the new value" << endl;
      exit(1);
    }	    
    patch_game(args[0],args[1],args[2]);
  }  else
    usage();
}

int main(int argc,char **argv)
{
  cout << "patlba v0.1 bugs to n3liw+@cmu.edu" << endl;
  add_vars();
  if(argc<2)
	  usage();
  scan_dir();
  argv++;				/*skip the program name*/
  parse_cmd(*argv,argv+1,argc-2);
  return 0;
}
